package be.cylab.java.wowa.training;

import info.debatty.java.aggregation.WOWA;

import java.util.Arrays;
import java.util.List;
import java.util.Random;

/**
 * Class of Solution. Solution with distance performance evaluation criterion
 */
public class SolutionDistance
        implements Comparable<SolutionDistance>, Cloneable {

    private double[] weights_w;
    private double[] weights_p;
    private double distance = Double.POSITIVE_INFINITY;

    /**
     * SolutionDistance constructor. Needs weight number as parameter
     *
     * @param weights_number
     */
    public SolutionDistance(final int weights_number) {
        this.weights_w = new double[weights_number];
        this.weights_p = new double[weights_number];
        for (int i = 0; i < weights_number; i++) {
            this.weights_w[i] = Math.random();
            this.weights_p[i] = Math.random();
        }
        this.normalize();
    }

    /**
     * Constructor solution distance. Made for test to initialize the Random
     * Generator
     *
     * @param weights_number int
     * @param seed           int
     */
    public SolutionDistance(final int weights_number, final int seed) {
        Random rnd = new Random(seed);
        this.weights_w = new double[weights_number];
        this.weights_p = new double[weights_number];
        for (int i = 0; i < weights_number; i++) {
            this.weights_w[i] = rnd.nextDouble();
            this.weights_p[i] = rnd.nextDouble();
        }
        this.normalize();
    }

    /**
     * @return
     */
    @Override
    public final String toString() {
        return "SolutionDistance{"
                + "weights_w=" + Arrays.toString(weights_w)
                + ", weights_p=" + Arrays.toString(weights_p)
                + ", distance=" + distance
                + '}';
    }

    /**
     * @param data
     * @param expected
     */
    final void computeScoreTo(final List<double[]> data,
                              final double[] expected
    ) {

        this.distance = 0;
        for (int i = 0; i < data.size(); i++) {
            double[] vector = data.get(i);
            double target_value = expected[i];
            WOWA wowa = new WOWA(this.weights_w, this.weights_p);
            double aggregated_value = wowa.aggregate(vector);
            this.distance += Math.pow(target_value - aggregated_value, 2);
        }
        this.distance = Math.sqrt(this.distance);

    }

    /**
     * @param probability
     */
    final void randomlyMutateWithProbability(final double probability) {
        double tos = Math.random();
        if (tos > probability) {
            // do nothing
            return;
        }

        //new weight value and position
        double new_weight_value = Math.random();
        int random_index = Utils.randomInteger(0, this.weights_p.length - 1);
        //Select w or p weights
        int weight_selection = Utils.randomInteger(0, 1);
        if (weight_selection == 0) {
            this.weights_w[random_index] = new_weight_value;
        } else {
            this.weights_p[random_index] = new_weight_value;
        }
        this.normalize();

    }

    /**
     * @param solution
     * @return int
     */
    @Override
    public final int compareTo(final SolutionDistance solution) {
        if (this.getDistance() > solution.getDistance()) {
            return 1;
        } else if (this.getDistance() < solution.getDistance()) {
            return -1;
        } else {
            return 0;
        }
    }

    /**
     *
     */
    final void normalize() {
        this.weights_w = Utils.normalizeWeights(this.weights_w);
        this.weights_p = Utils.normalizeWeights(this.weights_p);
    }

    /**
     * Copy.
     *
     * @return
     * @throws java.lang.CloneNotSupportedException if clone is not supported
     */
    public final SolutionDistance clone() throws CloneNotSupportedException {
        return (SolutionDistance) super.clone();
    }

    /**
     * @return
     */
    final double getDistance() {
        return distance;
    }

    final double[] getWeightsW() {
        return this.weights_w;
    }

    final double[] getWeightsP() {
        return this.weights_p;
    }
}
